local STATE_MODULE = "state"
local MODULE_MODULE = "module"

P2P_HOST, P2P_PORT, API_PORT, INIT_HOST, INIT_PORT = ...

if not P2P_HOST or not P2P_PORT or not API_PORT then
  print("Usage: moonbridge main/main.lua <P2P_HOST> <P2P_PORT> <API_PORT> [<INIT_HOST> <INIT_PORT>]")
  os.exit()
end

package.path = "main/?.lua;lib/?.lua;" .. package.path
package.cpath = "lib/extos/?.so;lib/json/?.so;" .. package.cpath

local container = require "container"

os.execute("mkdir -p socket")

container.load{ STATE_MODULE, MODULE_MODULE }

local modules = container.call_container_func("module", "get_module_list")

container.load(modules)

local function main()
  moonbridge_io.catch_sigterm()
  while true do
    local term = container.next_loop()
    if term then
      return
    end
  end
end

listen{
  { proto = "main" },
  pre_fork = 1, -- number of forks
  min_fork = 1, -- number of forks
  max_fork = 1, -- number of forks
  fork_delay = 0.25, -- seconds
  fork_error_delay = 2, -- seconds
  exit_delay = 60, -- seconds
  idle_timeout = 0, -- seconds (0 for no timeout)
  memory_limit = 128*1024*1024, -- bytes
  connect = main
}
