local json = require "json"
local moonbridge_io = require "moonbridge_io"
local http = require "moonbridge_http"

local sandbox = {}

function sandbox.new()

  local metatables = {} -- TODO weak table

  local function safe_table_wrap(func)
    return function(tbl, ...)
      if
        type(tbl) == "table" and
        metatables[tbl] == nil and
        debug.getmetatable(tbl) ~= nil
      then
        error("Security violation", 2)
      end
      return func(tbl, ...)
    end
  end

  local sandbox_G
  
  sandbox_G = {
    _VERSION = _VERSION,
    _G = sandbox_G,
    P2P_HOST = P2P_HOST,
    P2P_PORT = P2P_PORT,
    API_PORT = API_PORT,
    INIT_HOST = INIT_HOST,
    INIT_PORT = INIT_PORT,
    assert = assert,
    coroutine = coroutine,
    error = error,
    getmetatable = function(tbl)
      local mt = metatables[tbl]
      if mt ~= nil then
        local mt2 = rawget(mt, "__metatable")
        if mt2 then return mt2 else return mt end
      end
      if getmetatable(tbl) then return true end
      return nil
    end,
    http = http,
    ipairs = ipairs,
    json = {
      import = json.import,
      export = json.export,
      object = json.object,
      array = json.array,
      null = json.null,
      type = json.type
    },
    listen = listen,
    load = function(chunk, chunkname, mode, env)
      if mode ~= nil and mode ~= "t" then
        local funcname = debug.getinfo(1, "n").name or "load"
        error("bad argument #3 to '" .. funcname .. "' (string \"t\" or nil expected)", 2)
      end
      return load(
        chunk, chunkname, "t",
        env ~= nil and env or sandbox_G
      )
    end,
    -- TODO safety
    localconnect = moonbridge_io.localconnect,
    locallisten = moonbridge_io.locallisten,
    -- TODO safety
    next = next,
    --next = safe_table_wrap(next),
    os = {
      date = date,
      difftime = difftime,
      time = time
    },
    pairs = pairs, --[[function(tbl)
      local a, b, c = pairs(tbl)
      if
        a == next and
        (metatables[b] ~= nil or debug.getmetatable(b) == nil)
      then 
        return safe_next, b, c
      else
        local function returner(...)
          c = ...
          return ...
        end
        return function(b, c)
          return returner(a(b, c))
        end
      end
    end,--]]
    pcall = pcall,
    print = function(...)
      if string.match(({...})[1], "^%[") then
        return
      end
      local time = string.format("%018.8f", moonbridge_io.timeref())
      io.stderr:write(time, " ")
      io.stderr:write(...)
      io.stderr:write("\n")
    end,
    rawequal = rawequal,
    rawget = safe_table_wrap(rawget),
    rawlen = rawlen,
    rawset = safe_table_wrap(rawset),
    select = select,
    setmetatable = function(tbl, mt)
      local funcname = debug.getinfo(1, "n").name or "setmetatable"
      if type(tbl) ~= "table" then
        error("bad argument #1 to '" .. funcname .. "' (table expected)", 2)
      end
      if mt ~= nil then
        if type(mt) ~= "table" then
          error("bad argument #2 to '" .. funcname .. "' (nil or table expected)", 2)
        end
        if rawget(mt, "__gc") ~= nil then
          error("bad argument #2 to '" .. funcname .. "' (metatable must not contain __gc field)", 2)
        end
      end
      setmetatable(tbl, mt)
      metatables[tbl] = mt
      return tbl
    end,
    tcplisten = moonbridge_io.tcplisten,
    tcpconnect = moonbridge_io.tcpconnect_nb,
    time = moonbridge_io.timeref,
    os_time = os.time,
    tonumber = tonumber,
    tostring = tostring,
    type = type,
    xpcall = xpcall
  }

  for i, module in ipairs{"math", "string", "table", "utf8"} do
    sandbox_G[module] = {}
    for key, value in pairs(_G[module]) do
      local t = type(value)
      if t=="function" or t=="string" or t=="number" then
        sandbox_G[module][key] = value
      end
    end
  end
  sandbox_G.math.randomseed = nil

  return sandbox_G
end

return sandbox
