local api_module = {
  ident = "api",
  uses = {
    "api"
  },
  need_single = {
    "person",
    "user",
    "pki",
    "module",
    "id_generator"
  },
  provide = {
    api = {},
  }
}

function api_module.provide.api.file(args)
  if files[args.path] then
    return json.object{
      content_type = "text/html",
      data = files[path].content
    }
  else
    return json.object{
      status = 404
    }
  end
end

function api_module.provide.api.get_modules(args)
end

---
local api = {}

local layout = [[
  <html><head>
    <title>LiquidFeedback Blockchain prototype 1</title>
    <link rel="stylesheet" type="text/css" href="/api/style.css">
    <script src="/api/api.js"></script>
    #{html_head}
  </head><body>
    <div class="page">
      <h1>LiquidFeedback Blockchain</h1>
      <div class="template__content">
        #{content}
      </div>
      <div class="push"></div>
    </div>
    <footer>
      <div>
        <span>Applications:</span>
        <ul id="template__applications"></ul>
      </div>
      <div>
        <span>Settings: </span>
        <ul id="template__containers"></ul>
      </div>
      <div>
        <div>
          Node ID: #{node_id} &nbsp; 
          User ID: #{user_id} &nbsp; 
          Person ID: #{person_id} &nbsp;
          <a href="#" onclick="logout(); return false;">Logout</a>
        </div>
      </div>
    </footer>
    <script>
      function logout() {
        api("POST", "api", { command: "drop_session" }, function(result) {
          window.location.reload()
        })
      }
      api("POST", "api", { command: "get_containers", with_frontend: true }, function(result) {
        for (container_id in result) {
          var container = result[container_id];
          var a_el = document.createElement("a");
          a_el.innerHTML = container_id;
          a_el.href = "/" + container_id + "/";
          var li_el = document.createElement("li");
          li_el.appendChild(a_el);
          if (container.is_application) {
            document.getElementById("template__applications").appendChild(li_el);
          } else {
            document.getElementById("template__containers").appendChild(li_el);
          }
        }
      })
    </script>
  </body></html>
]]

local files = {
  [""] = {
    content = [[API is running.]]
  },
  ["api.js"] = {
    data = [[
      function api(method, module, data, next) {
        var xHttp = new XMLHttpRequest();
        xHttp.onreadystatechange = function() {
          if (this.readyState != 4) return;
          if (this.status != 200) {
            console.log("Request failed.");
            return;
          }
          var result = JSON.parse(this.responseText);
          if (result) {
            if (next) {
              next(result);
            }
          } else {
            alert("Invalid API response:\n" + this.responseText);
          }
        };
        xHttp.open(method, "/" + module + "/api");
        xHttp.setRequestHeader("Content-type", "application/x-www-form-urlencoded")
        xHttp.send(JSON.stringify(data));
      }

      function addTransaction(module, data, next) {
        var transaction = { module: module, data: data }
        api("POST", "blockchain", { command: "new_transaction", transaction: transaction }, next);
      }
    ]]
  },
  ["style.css"] = {
    data = [[
      body { margin: 0; padding: 0; font-family: sans-serif; min-height: 100%;} 
      h1 { background: rgb(76,175,80); color: #fff; padding: 5px 20px; margin: 0; font-size: 12pt; }
      .template__content { margin: 20px; }
      footer { height: 60px; background: #444; padding: 20px; line-height: 166%; font-size: 90%; position: relative; z-index: 2;}
      footer, footer a { color: #fff; }
      footer ul { display: inline-block; margin: 0; padding: 0; }
      footer ul li { display: inline-block; margin: 0; padding: 0; margin-left: 1em; font-weight: bold;}
      label { display: block; text-transform: uppercase; font-size: 90%; font-weight: bold; color: #444; }
      input[type=text], textarea { display: block; margin-bottom: 2ex; width: 100%; }
      input[type=checkbox] + label { display: inline }
      textarea { height: 20ex; }
      .hidden { display: none; }
      .page {
        min-height: 100%;
        height: auto !important;
        height: 100%;
        margin: 0 auto -100px;
        xmin-width: 912px;
        position: relative;
        overflow: hidden;
        z-index: 1;
      }
      .push {
        height: 100px;
      }
      
    ]]
  }
}

local sessions = {}

function api.has_frontend()
  return true
end

function api_module.provide.api.on_api_get_file(args)
  print("[API] on api get file '" .. args.file .. "'")
  local file = files[args.file]
  if file and file.content then
    file.content = string.gsub(file.content, "#{user_id}", tostring(args.user_id))
    file.content = string.gsub(file.content, "#{person_id}", tostring(args.person_id))
  end
  return file
end

function api_module.provide.api.on_api_call(args)
  local method = args.method
  local data = args.data
  local command = data.command
  if command == "get_containers" then
    local result = {}
    for m in pairs(module.api) do
      result[m] = json.object{
        is_application = module.api[m].is_application and module.api[m].is_application() or false
      }
    end
    return result
  elseif command == "drop_session" then
    sessions[args.session_id] = nil
    return json.object()
  end
end

function api_module.init()
  spawn(function()
    print("[API/init] started")

    local listen = locallisten("socket/socket-api")
    print("[API/init] listener installed")
    
    print("[API/init] starting mainloop")
    while true do
      
      local socket
      local first = true
      repeat
        coroutine.yield(listen, "r", first, "moonbridge_io")
        first = false
        socket = listen:accept_nb()
      until socket
      
      print("[API/loop] accepted socket from API/server")
      spawn(function()
        print("[API/handler] new socket handler for API/server started, reading request from socket")
        local data = socket:read_yield(nil, "\n")
        local request = json.import(data)
        print("[API/handler] request: " .. json.export(request))
        if request.command == "authenticate_with_password" then
          print("[API/handler] authenticate with password")
          local result = json.object()
          local r_auth = module.user.authenticate_with_password{
            login = request.login,
            password = request.password
          }
          local user_id = r_auth.user_id
          if user_id then
            local r_id = module.id_generator.create_id()
            local session_id = r_id.id
            local session = json.object{
              user_id = user_id,
              person_id = r_auth.person_id
            }
            sessions[session_id] = session
            result.session_id = session_id
            result.user_id = user_id
            result.person_id = r_auth.person_id
            result.admin = r_auth.admin
            result.home_container = r_auth.home_container
          end
          socket:flush(json.export(result))
        elseif request.command == "get_session" then
          print("[API/handler] get session")
          socket:write(json.export(sessions[request.session_id]))
        elseif request.command == "get_module_file" then
          if module.api[request.module] then
            print("[API/handler] serving module file: " .. request.module .. " " .. request.file)
            local result = module.api[request.module].on_api_get_file{ file = request.file, user_id = request.user_id, person_id = request.person_id }
            if result then
              socket:write(json.export(result))
            end
          end
        elseif request.command == "call_module" then
          print("[API/handler] serving module api call: " .. request.module .. " " .. json.export(request.data))
          print("[API/handler] calling on_api_call of module: " .. request.module)
          local result = module.api[request.module].on_api_call{ method = request.method, data = request.data, user_id = request.user_id, person_id = request.person_id, session_id = request.session_id }
          if result then
            print("[API/handler] got result from module: " .. request.module)
            socket:write(json.export(result))
          else
            print("[API/handler] no result from module: " .. request.module)
          end
        end
        print("[API/handler] sending result to API/serve")
        socket:flush("\n")
        socket:close()
      end)
      
    end
  end )
end

local function ask(command)
  local socket, err_message = localconnect("socket/socket-api")
  socket:flush(json.export(json.object(command)), "\n")
  local r = socket:read(nil, "\n")
  local result
  if r then
    result = json.import(r)
  end
  if result and type(result) == "table" then
    return result
  else
    return json.object()
  end
end

local function serve(request)
  print("[API/serve] " .. request.method .. " /" .. request.path)
  local result
  local set_cookie
  
  local session_id
  local user_id
  local person_id
  
  local cookies = request.headers["Cookie"]
  for i, cookie in ipairs(cookies) do
    session_id = string.match(cookie, "^lfb_session=([^;]+)")
  end

  if session_id then
    local r = ask{
      command = "get_session",
      session_id = session_id
    }
    if r.user_id then
      user_id = r.user_id
      person_id = r.person_id
    end
  end

  if not user_id then
    if request.method == "POST" then
      local r_auth = ask{
        command = "authenticate_with_password",
        login = request.post_params.login,
        password = request.post_params.password
      }
      if r_auth and r_auth.user_id then
        local redirect_module = "api"
        if r_auth.home_container then
          redirect_module = r_auth.home_container
        end
        print("[API/serve] redirecting to module " .. redirect_module)
        request:send_status("302 redirect")
        request:send_header("Set-Cookie", "lfb_session=" .. r_auth.session_id .. "; path=/")
        request:send_header("Location", "/" .. redirect_module .. "/")
        request:finish()
        return
      end
    end
    result = [[
      <html><head>
        <title>Login</title>
      </head><body>
        <h1>Login</h1>
        <form method="post">
          <input type="text" name="login" placeholder="login"><br /><br />
          <input type="password" name="password" placeholder="password"><br /><br />
          <input type="submit" value="login">
        </form>
      </body><html>
    ]]
  end

  if user_id then
    local data = json.object()
    if request.path == "" then
      if request.method == "GET" then
        print("[API/serve] redirecting to module api")
        request:send_status("302 redirect")
        request:send_header("Location", "/api/")
        request:finish()
        return
      end
    elseif request.path == "favicon.ico" then
    elseif string.match(request.path, "^[^/]+$") then
      if request.method == "GET" then
        local module = string.match(request.path, "^([^/]+)$")
        print("[API/serve] redirecting to module frontend: " .. module)
        request:send_status("302 redirect")
        request:send_header("Location", "/" .. module .. "/")
        request:finish()
        return
      end
    elseif string.match(request.path, "^[^/]+/api$") then
      local module = string.match(request.path, "^([^/]+)/api$")
      local module_request = json.import(request.body)
      print("[API/serve] serving module api request: " .. module .. " " .. json.export(module_request))
      data.command = "call_module"
      data.module = module
      data.method = request.method
      data.data = module_request
      data.user_id = user_id
      data.person_id = person_id
      data.session_id = session_id
    elseif string.match(request.path, "^[^/]+/") then
      local module, file = string.match(request.path, "^([^/]+)/(.*)")
      if request.method == "GET" then
        print("[API/serve] serving module file: " .. module .. " " .. file)
        data.command = "get_module_file"
        data.module = module
        data.file = file
        data.user_id = user_id
        data.person_id = person_id
      end
    end

    if data.command then
      data = json.export(data)
      local socket, err_message = localconnect("socket/socket-api")
      if socket then
        socket:flush(data, "\n")
        result = socket:read()
        socket:close()
      else
        result = err_message
      end
    end

    if string.match(request.path, "^[^/]+/") then
      if request.method == "GET" then
        local result_obj = json.import(result)
        if result_obj.data then
          result = result_obj.data
        else
          result_obj.user_id = user_id
          result_obj.node_id = node_id
          result_obj.person_id = person_id
          result = string.gsub(layout, "#{([^}]+)}", function(match)
            return result_obj[match] or ""
          end)
        end
      end
    end
    if not result or result == "\n" then
      request:send_status("404 Not Found")
      request:send_data("<h1>404 Not Found</h1>")
      request:finish()
      return
    end
  end
  request:send_status("200 OK")
  request:send_data(result)
  request:finish()
end

listen{
  { proto = "tcp", host = "127.0.0.1", port = API_PORT },
  pre_fork = 2, -- number of forks
  min_fork = 2, -- number of forks
  max_fork = 5, -- number of forks
  fork_delay = 0.25, -- seconds
  fork_error_delay = 2, -- seconds
  exit_delay = 60, -- seconds
  idle_timeout = 0, -- seconds (0 for no timeout)
  memory_limit = 128*1024*1024, -- bytes
  
  connect = http.generate_handler(
    {
      body_size_limit = 32*1024*1024
    },
    function(request)
      local errmsg
      local status, result = xpcall(
        serve,
        function(err)
          errmsg = debug.traceback(
            "Unexpected error in request handler: " .. tostring(err),
            2
          )
        end,
        request
      )
      if errmsg then
        io.stderr:write(errmsg)
        local status, result = pcall(function()
          request:send_status("500 Internal Server Error")
          request:send_header("Content-Type", "text/html")
          request:send_data("<html>\n<head><title>500 Internal Server Error</title></head>\n<body><h1>500 Internal Server Error</h1></body>\n</html>\n")
          request:finish()
        end)
        if not status then
          io.stderr:write("Could not send error page: " .. tostring(result))
        end
        return false
      end
      return true
    end
  ),

  -- executed on process termination
  finish = function()
  end
}

return api_module
