local crypto_module = {
  ident = "crypto",
  provide = {
    pki = {},
    hash = {},
    id_generator = {}
  }
}
-- ------------------------------------------------------------------------
-- FRONTEND
-- ------------------------------------------------------------------------

local files = {
  [""] = {
    content = [[
      <h2>Module: Crypto</h2>
      <div>Info:</div>
      <div id="info">
      </div>
      <script>
        api("POST", "crypto", { command: "get_info" }, function(result) {
          document.getElementById("info").innerHTML = "";
          for (key in result) {
            var li_el = document.createElement("li");
            li_el.innerHTML = key + ": " + result[key]
            document.getElementById("info").appendChild(li_el);
          }
        })
      </script>
    ]]
  }
}

function crypto_module.has_frontend()
  return true
end

function crypto_module.on_api_call(args)
  local method = args.method
  local data = args.data
  local command = data.command
  if command == "get_info" then
    return {
      keys = "TODO",
      signature = "TODO",
      hash = "sha256"
    }
  end
end

function crypto_module.on_api_get_file(args)
  print("[Crypto] on api get file '" .. args.file .. "'")
  return files[args.file]
end

local function hash(data)
  local hash, err, status = extos.pfilter(data, "openssl", "dgst", "-sha256")
  local hash = string.match(hash, " (.+)\n")
  return hash
end

function crypto_module.provide.hash.hash_object(args)
  --local obj = args.obj
  --local old_hash = obj.hash
  --obj.hash = nil
  return hash(json.export(args.obj))
  --[[
  if old_hash then
    obj.hash = old_hash
    return obj, hash == old_hash
  else
    obj.hash = hash
  end
  return obj, true
  --]]
end

function crypto_module.provide.id_generator.create_id()
  local id, err, status = extos.pfilter(data, "openssl", "rand", "-hex", "16")
  local id = string.match(id, "(.+)\n")
  return json.object{ id = id }
end

function crypto_module.provide.pki.create_key_pair()
  local private_key, err, status = extos.pfilter(nil, "openssl", "genpkey", "-algorithm", "RSA", "-pkeyopt", "rsa_keygen_bits:2048")
  local public_key, err, status = extos.pfilter(private_key, "openssl", "rsa", "-pubout")
  local fingerprint = hash(public_key)
  
  return {
    private_key = private_key,
    public_key = public_key,
    fingerprint = fingerprint
  }
end

function crypto_module.provide.pki.sign(args)
  local data = args.data
  local private_key = args.private_key
  
  local tmp = crypto_module.provide.id_generator.create_id()
  
  local fh = io_open("tmp/" .. tmp.id, "w")
  fh:write(private_key)
  fh:close()
  
  local signature, err, status = extos.pfilter(data, "openssl", "dgst", "-sha256", "-sign", "tmp/" .. tmp.id)

  local signature_base64, err, status = extos.pfilter(signature, "openssl", "base64")
  
  return json.object{
    signature = signature_base64
  }
end

function crypto_module.provide.pki.check_signature(args)
  local public_key = args.public_key
  local obj = args.obj
  local signature = obj.signature
  obj.signature = nil
  local data = json.export(obj)
  
  local tmp_key = crypto_module.provide.id_generator.create_id()
  local fh = io_open("tmp/" .. tmp_key.id, "w")
  fh:write(public_key)
  fh:close()
  
  local tmp = crypto_module.provide.id_generator.create_id()
  local dummy, err, status = extos.pfilter(signature, "openssl", "base64", "-d", "-out", "tmp/" .. tmp.id)
  local result, err, status = extos.pfilter(data, "openssl", "dgst", "-sha256", "-verify", "tmp/" .. tmp_key.id, "-signature", "tmp/" .. tmp.id)
  
  local valid = false
  if result == "Verified OK\n" then
    valid = true
  end
  obj.signature = signature  
  return json.object{
    valid = valid
  }
end

return crypto_module
