local demo01_module = {
  ident = "demo01",
  need_single = {
    "hash",
    "state",
    "person"
  },
  provide = {
    api = {},
    app = {}
  }
}


local state

local files = {
  [""] = {
    content = [[
      <h2>Demo application 1</h2>
      <button onclick="document.getElementById('new_initiative_form').classList.toggle('hidden'); return false;">Create new initiative</button>
      <br /><br />
      <form id="new_initiative_form" class="hidden">
        <div>
          <label>Title</label>
          <input type="text" name="title">
        </div>
        <div>
          <label>Text</label>
          <textarea name="text"></textarea>
        </div>
        <div>
          <label>Changes</label>
          <textarea name="changes">{}</textarea>
        </div>
        <div>
          <input type="submit" value="create new initiative" onclick="newInitiative(this.form); return false;">
        </div>
      </form>
      <h3>State:</h3>
      <div id="state"></div>
      <h3>Open initiatives:</h3>
      <div id="open_initiatives"></div>
      <h3>Accepted initiatives:</h3>
      <div id="accepted_initiatives"></div>
      <h3>Rejected initiatives:</h3>
      <div id="rejected_initiatives"></div>
      <script>
        function newInitiative(form) {
          addTransaction("demo01", {
            intent: "new_initiative",
            title: form.title.value,
            text: form.text.value,
            changes: JSON.parse(form.changes.value)
          }, function() {
            form.reset();
            document.getElementById('new_initiative_form').classList.add('hidden');
            updateInitiatives();
          });
        }
        function vote(initiative_id, decision) {
          addTransaction("demo01", {
            intent: "vote",
            initiative_id: initiative_id,
            decision: decision
          }, function() {
            updateState();
            updateInitiatives();
          });
        }
        function recurseState(state, depth = 0) {
          var html = "<ul>"
          for (var key in state) {
            var value = state[key];
            if (typeof value == "object") {
              if (!(depth == 0 && (key == "initiatives" || key == "nodes"))) {
                html += "<li>" + key;
                html += recurseState(value, depth + 1);
                html += "</li>";
              }
            } else {
              html += "<li>" + key + " = " + value + "</li>"
            }
          }
          html += "</ul>"
          return html;
        };
        function updateState() {
          api("POST", "demo01", { command: "get_state" }, function(result) {
            document.getElementById("state").innerHTML = recurseState(result);
          });
        }
        function updateInitiatives() {
          api("POST", "demo01", { command: "get_initiatives" }, function(result) {
            document.getElementById("open_initiatives").innerHTML = "";
            document.getElementById("accepted_initiatives").innerHTML = "";
            document.getElementById("rejected_initiatives").innerHTML = "";
            for (initiative_id in result) {
              var initiative = result[initiative_id];
              var li_el = document.createElement("li");
              li_el.innerHTML = initiative_id + " = " + initiative.title
              if (!initiative.accepted && !initiative.rejected) {
                li_el.innerHTML += " <button onclick='vote(\"" + initiative_id + "\", \"yes\");return false;'>YES</button> <button onclick='vote(\"" + initiative_id + "\", \"no\"); return false;'>NO</button>";
              };
              li_el.innerHTML += "<br />" + JSON.stringify(initiative.changes)
              if (initiative.votes) {
                var inner_ul_el = document.createElement("ul");
                for (person_id in initiative.votes) {
                  var inner_li_el = document.createElement("li");
                  inner_li_el.innerHTML = person_id + " = " + initiative.votes[person_id];
                  inner_ul_el.appendChild(inner_li_el);
                }
                li_el.appendChild(inner_ul_el);
              }
              var elId = "open_initiatives";
              if (initiative.accepted) {
                elId = "accepted_initiatives";
              } else if (initiative.rejected) {
                elId = "rejected_initiatives";
              }
              document.getElementById(elId).appendChild(li_el);
                
            }
          });
        }
        updateState();
        updateInitiatives();
      </script>
    ]]
  }
}

function demo01_module.provide.api.is_application()
  return true
end

function demo01_module.provide.api.on_api_get_file(args)
  print("[demo01] on api get file '" .. args.file .. "'")
  return files[args.file]
end

function demo01_module.provide.api.on_api_call(args)
  print("[demo01] on api call '" .. tostring(args.command) .. "'")
  local method = args.method
  local data = args.data
  local command = data.command
  if command == "get_state" then
    return state.state
  elseif command == "get_initiatives" then
    return state.initiatives
  end
end

function demo01_module.provide.app.process_transaction(args)
  local transaction = args.transaction
  if transaction.data.intent == "new_initiative" then
    local initiative = json.object{
      title = transaction.data.title,
      text = transaction.data.text,
      changes = transaction.data.changes,
      votes = json.object()
    }
    local initiative_hash = module.hash.hash_object{ obj = initiative }
    state.initiatives[initiative_hash] = initiative
  elseif transaction.data.intent == "vote" then
    local initiative = state.initiatives[transaction.data.initiative_id]
    if initiative then
      initiative.votes[transaction.person_id] = transaction.data.decision
    end
    local population_count = module.person.get_count()
    local needed_count = population_count / 2
    for initiative_id, initiative in pairs(state.initiatives) do
      local yes_count = 0
      local no_count = 0
      for person_id, decision in pairs(initiative.votes) do
        if decision == "yes" then
          yes_count = yes_count + 1
        elseif decision == "no" then
          no_count = no_count + 1
        end
      end
      if yes_count > needed_count then
        initiative.accepted = true
        if initiative.changes then
          for i, change in pairs(initiative.changes) do
            local current_path = state.state
            local path = {}
            if change.key then
              for path_element in string.gmatch(change.key, "[^.]+") do
                table.insert(path, path_element)
              end
              for i, path_element in ipairs(path) do
                if i < #path then
                  if type(current_path[path_element]) ~= "table" then
                    current_path[path_element] = json.object()
                  end
                  current_path = current_path[path_element]
                else
                  if change.value ~= nil then
                    current_path[path_element] = change.value
                  else
                    current_path[path_element] = nil
                  end
                end
              end
            end
          end
        end
      elseif no_count > needed_count then
        initiative.rejected = true
      end
    end
  end
end

function demo01_module.init()
  spawn(function()
    print("[Demo01/init] started")
    state = module.state.get_state()
    if not state.initiatives then
      state.initiatives = json.object()
    end
    if not state.state then
      state.state = json.object()
    end
    print("[Demo01/init] finished")
  end)
end
  
return demo01_module
