local module_module = {
  ident = "module",
  need_single = {
    "state"
  },
  provide = {
    module = {},
  }
}
local module_list = { "state", "module", "user", "api", "p2p", "crypto", "blockchain", "node", "person", "consensus", "demo01" }

local state = json.object{
  containers = {
    state = { module_id = "state" },
    module = { module_id = "module", relations = json.object{ state = "state" } },
    api = { module_id = "api", relations = json.object{ module = "module", person = "person", crypto = "crypto" } },
    p2p = { module_id = "p2p", relations = json.object{ node = "node", random_node_chooser = "consensus", module = "module" } },
    crypto = { module_id = "crypto" },
    blockchain = { module_id = "blockchain", relations = json.object{ state = "state", crypto = "crypto" } },
    node = { module_id = "node", relations = json.object{ state = "state", crypto = "crypto" } },
    person = { module_id = "person", relations = json.object{ state = "state" } },
    consensus = { module_id = "consensus", relations = json.object{ random_node_chooser = "person", blockchain = "blockchain", crypto = "crypto" }, },
    demo01 = { module_id = "demo01", relations = json.object{ state = "state", crypto = "crypto" } }
  }
}

-- ------------------------------------------------------------------------
-- FRONTEND
-- ------------------------------------------------------------------------

local files = {
  [""] = {
    content = [[
      <h2>Module: module</h2>
      <form>
        <textarea name="config" id="config" style="width: 100%; height: 40ex;"></textarea>
        <button onclick="saveConfig(this.form.config.innerHTML); return false;">save config</button>
      </form>
      <br />
      <br />
      <div>Active containers:</div>
      <div id="containers"></div>
      <script>
        function saveConfig(config) {
          addTransaction("module", {
            intent: "config",
            config: JSON.parse(config)
          });
        }
        api("POST", "module", { command: "get_containers", with_relations: true }, function(result) {
          document.getElementById("config").innerHTML = JSON.stringify(result, null, 2);
          document.getElementById("containers").innerHTML = "";
          for (container_id in result) {
            var container = result[container_id];
            var outer_li_el = document.createElement("li");
            var div_el = document.createElement("div");
            div_el.innerHTML = container_id + " = " + container.module_id;
            outer_li_el.appendChild(div_el);
            if (container.relations) {
              var ul_el = document.createElement("ul");
              for (relation_id in container.relations) {
                container_id = container.relations[relation_id];
                var li_el = document.createElement("li");
                li_el.innerHTML = relation_id + " = " + container_id;
                ul_el.appendChild(li_el);
              }
              outer_li_el.appendChild(ul_el);
            }
            document.getElementById("containers").appendChild(outer_li_el);
          }
        })
      </script>
    ]]
  }
}

function module_module.has_frontend()
  return true
end

function module_module.on_api_get_file(args)
  print("[module] on api get file '" .. args.file .. "'")
  return files[args.file]
end

function module_module.on_api_call(args)
  local method = args.method
  local data = args.data
  local command = data.command
  if command == "get_containers" then
    local result = module_module.get_containers{ with_frontend = data.with_frontend }
    return result
  end
end

-- ------------------------------------------------------------------------
-- GET CONTAINERS
-- ------------------------------------------------------------------------

function module_module.get_module_list()
  return module_list
end

function module_module.provide.module.get_containers(args)
  local result
  if args and args.with_frontend then
    local containers = json.object{}
    for container_id, container in pairs(state.containers) do
      if module[container_id].call("has_frontend") then
        containers[container_id] = container
      end
      if module[container_id].call("is_application") then
        container.is_application = true
      end
    end
    result = containers
  else
    result = state.containers
  end
  return result
end

function module_module.get_related_container(args)
  local relations = state.containers[args.container_id].relations
  if relations then
    other_container_id = relations[args.relation]
    return other_container_id
  end
end

-- ------------------------------------------------------------------------
-- INIT
-- ------------------------------------------------------------------------

function module_module.init()
  print("[Module/init] started")
  print("[Module/init] preparing state")
  local containers = state.containers
  state = module.state.get_state()
  if not state.containers then
    state.containers = containers
  end
  print("[Module/init] finished")
end

return module_module
