local person_module = {
  ident = "person",
  need_single = {
    "state",
    "store",
    "pki",
    "user"
  },
  provide = {
    api = {},
    person = {},
    choose_node = {},
    node_config = {}
  },
  provide_single = {
    sync = {},
    app = {}
  }
}

local state
local store

-- ------------------------------------------------------------------------
-- FRONTEND
-- ------------------------------------------------------------------------

local files = {
  [""] = {
    content = [[
      <h2>Module: person</h2>
      
      <h3>Update node config</h3>
      <form id="update_node_config_form">
        <label>Node ID:</label>
        <input type="text" name="node_id">
        <input type="submit" value="OK" onclick="updateNodeConfig(this.form); return false;">
      </form>
      
      <div>Accreditation requests:</div>
      <div id="persons"></div>
      <div>Accredited persons:</div>
      <div id="accredited_persons"></div>
      
      <script>
        function updateNodeConfig(form) {
          api("POST", "person", { command: "update_node_config", data: {
            node_id: form.node_id.value,
          } }, function(result) {
            document.getElementById('update_node_config_form').reset();
            //document.getElementById('add_user_form').classList.add('hidden');
            updatePersons()
          });
        }
        function updatePersons() {
          api("POST", "person", { command: "get_persons" }, function(result) {
            console.log("persons", result);
            document.getElementById("persons").innerHTML = "";
            document.getElementById("accredited_persons").innerHTML = "";
            for (person_id in result) {
              var person = result[person_id];
              var li_el = document.createElement("li");
              if (person.accredited) {
                li_el.innerHTML = person_id + " = " + person.identification;
                if (person.node_config) {
                  li_el.innerHTML += " " + JSON.stringify(person.node_config);
                }
                document.getElementById("accredited_persons").appendChild(li_el);
              } else {
                li_el.innerHTML = person_id;
                var inner_ul = document.createElement("ul")
                li_el.appendChild(inner_ul);
                for (identification in person.votes) {
                  var inner_li = document.createElement("li")
                  inner_li.innerHTML = identification
                  inner_ul.appendChild(inner_li);
                }
                document.getElementById("persons").appendChild(li_el);
              }
            }
          })
        }
        updatePersons();
      </script>
    ]]
  },
  ["accredit"] = {
    content = [[
      <h2>Module: person</h2>
      <form id="accredit_person_form">
        <div>
          <label>ID</label>
          <input type="text" name="person_id" id="person_id">
        </div>
        <div>
          <label>Identification</label>
          <input type="text" name="identification">
        </div>
        <div>
          <input type="submit" value="accredit person" onclick="accreditPerson(this.form); return false;">
        </div>
      </form>
      
      <script>
        function accreditPerson(form) {
          addTransaction("person", {
            intent: "accredit",
            person_id: form.person_id.value,
            identification: form.identification.value,
            decision: "yes"
          });
          window.location.replace(".")
        }
        var tmp = /person_id=([^&]+)/.exec(window.location.href);
        if (tmp) {
          var accredit_person_id = tmp[1];
          document.getElementById("person_id").value = accredit_person_id;
        }
      </script>
    ]]
  }
}

function person_module.provide.api.on_api_get_file(args)
  print("[Person] on api get file '" .. args.file .. "'")
  return files[args.file]
end

function person_module.provide.api.on_api_call(args)
  local method = args.method
  local data = args.data
  local command = data.command
  if command == "update_node_config" then
    if data.data then
      local person_id = args.person_id
      local node_id = data.data.node_id
      if not store.persons[person_id] then
        store.persons[person_id] = json.object()
      end
      local public_key = module.user.get_public_key{ user_id = args.user_id }
      local index = store.persons[person_id].node_config and store.persons[person_id].node_config.index or 0
      local node_config = json.object{
        public_key = public_key,
        node_id = node_id,
        index = index + 1
      }
      node_config.signature = module.user.sign_transaction{ transaction = node_config, user_id = args.user_id }
      store.persons[person_id].node_config = node_config
      print("[Person] set node for " .. person_id .. ": " .. node_id)
    end
    return { state = "ok" }
  elseif command == "get_persons" then
    return state.persons
  end
end


-- ------------------------------------------------------------------------
-- AUTHENTICATE PERSON
-- ------------------------------------------------------------------------

function person_module.provide.person.authenticate_with_password(args)
  if logins[args.login] and args.password == logins[args.login].password then
    return json.object{ 
      person_id = logins[args.login].person_id,
      admin = logins[args.login].admin,
      home_container = logins[args.login].home_container
    }
  end
  return json.object{ message = "Invalid login or password" }
end

function person_module.provide.person.get_identification(args)
  if not state.persons then
    return
  end
  return state.persons[args.person_id]
end

function person_module.provide.person.get_count()
  local person_count = 0
  for person_id in pairs(state.persons) do
    person_count = person_count + 1
  end
  return person_count
end

-- ------------------------------------------------------------------------
-- RANDOM NODE CHOOSER
-- ------------------------------------------------------------------------

function person_module.provide.choose_node.get_random_node_ids(args)
  local node_ids = {}
  local person_count = 0
  if not state.persons then
    return node_ids
  end
  for person_id in pairs(state.persons) do
    person_count = person_count + 1
  end
  for i = 1, args.count do
    local r = math.floor(math.random() * person_count + 1)
    local node_id
    for person_id, person in pairs(state.persons) do
      if r == 1 and store.persons[person_id] then
        local node_config = store.persons[person_id].node_config
        if node_config and node_config.node_id then
          node_id = node_config.node_id
        end
      end
      r = r - 1 
    end
    if node_id then
      node_ids[node_id] = (node_ids[node_id] or 0) + 1
    end
  end
  return node_ids
end

-- ------------------------------------------------------------------------
-- TRANSACTION HANDLER
-- ------------------------------------------------------------------------

function person_module.provide_single.app.process_transaction(args)
  local transaction = args.transaction
  if not transaction then
    return
  end
  if transaction.module ~= "person" then
    return
  end
  if transaction.data.intent == "accredit" then
    if not state.persons then
      if transaction.person_id == transaction.data.person_id then
        state.persons = json.object()
        state.persons[transaction.data.person_id] = json.object{
          id = transaction.data.person_id,
          accredited = true,
          identification = transaction.data.identification,
          votes = json.object{}
        }
      end
      return
    end
    if not state.persons[transaction.data.person_id] then
      state.persons[transaction.data.person_id] = json.object{
        id = transaction.data.person_id,
        votes = json.object{}
      }
    end
    if not state.persons[transaction.data.person_id].votes then
      state.persons[transaction.data.person_id].votes = json.object()
    end
    if not state.persons[transaction.data.person_id].votes[transaction.data.identification] then
      state.persons[transaction.data.person_id].votes[transaction.data.identification] = json.object{}
    end
    local votes = state.persons[transaction.data.person_id].votes[transaction.data.identification]
    votes[transaction.person_id] = transaction.data.decision
    if not state.persons[transaction.data.person_id].accredited then
      local accredited_person_count = 0
      for person_id, person in pairs(state.persons) do
        if person.accredited then
          accredited_person_count = accredited_person_count + 1
        end
      end
      local votes_needed = math.ceil(accredited_person_count / 2 + 0.5)
      local accredited_as
      for identification, votes in pairs(state.persons[transaction.data.person_id].votes) do
        local yes = 0
        local no = 0
        for person_id, vote in pairs(votes) do
          if vote == "yes" then
            yes = yes + 1
          elseif vote == "no" then
            no = no + 1
          end
        end
        if yes >= votes_needed then
          accredited_as = identification
        end
      end
      if accredited_as then
        state.persons[transaction.data.person_id].accredited = true
        state.persons[transaction.data.person_id].identification = accredited_as
      end
    end
  end
end

-- ------------------------------------------------------------------------
-- SYNC
-- ------------------------------------------------------------------------


function person_module.has_sync()
  return true
end

local function merge_remote_persons(persons)
  for person_id, person in pairs(persons) do
    if person.node_config then
      local check = module.pki.check_signature{ obj = person.node_config, public_key = person.node_config.public_key }
      if check.valid == true then
        if not store.persons[person_id] then
          print(" [Person] added person: " .. person_id)
          store.persons[person_id] = json.object()
        end
        if not store.persons[person_id].node_config or person.node_config.index > store.persons[person_id].node_config.index then
          print(" [Person] added/updated node config: " .. person_id)
          store.persons[person_id].node_config = person.node_config
        end
      else
        print(" [Person/merge_remote_nodes] invalid signature, config dropped")
      end
    end
  end
end

function person_module.provide_single.sync.sync_in(args)
  print("[Person/sync_in]")
  local person_id = args.person_id
  local iteration = args.iteration
  local request = args.request
  local remote_persons = request.persons
  if remote_persons then
    merge_remote_persons(remote_persons)
  end
  if iteration == 1 then
    return json.object{ persons = store.persons }
  end
end

function person_module.provide_single.sync.sync_out(args)
  print("[Person/sync_out]")
  local person_id = args.person_id
  local iteration = args.iteration
  local result = args.result
  if result then
    local remote_persons = result.persons
    if remote_persons then
      merge_remote_persons(remote_persons)
    end
  end
  if iteration == 1 then
    return json.object{ persons = store.persons }
  end
end

function person_module.provide_single.sync.sync_out_completed()
end

function person_module.provide_single.sync.sync_out_cancelled()
end

-- ------------------------------------------------------------------------
-- INIT
-- ------------------------------------------------------------------------

function person_module.init()
  spawn(function()
    print("[Person/init] started")
    state = module.state.get_state()
    if not state.persons then
      state.persons = json.object()
    end
    store = module.store.get_store()
    if not store.persons then
      store.persons = json.object()
    end
    print("[Person/init] finished")
  end)
end
  
return person_module

