local state_module = {
  ident = "state",
  provide = {
    state = {},
    store = {},
    manage_state = {}
  }
}

local store

function state_module.provide.store.get_store()
  print("[State/get_store] for calling module " .. calling_module)
  if not store[calling_module] then
    store[calling_module] = json.object()
  end
  return store[calling_module]
end

function state_module.provide.store.save()
  print("[State/save_store]")
  local data = json.export(store)
  save_store(data)  
end

local state_index = 0
local state

function state_module.provide.state.get_state()
  print("[State/get_state] for calling module " .. calling_module)
  if not state[calling_module] then
    state[calling_module] = json.object()
  end
  return state[calling_module]
end

function state_module.provide.manage_state.set_index(args)
  local index = args.index
  print("[State/set_index] set index to " .. index .. " " .. json.export(state))
  local data = json.export(state)
  save_state(index, data)
  state_index = index
end

function state_module.provide.manage_state.revert(args)
  local index = args.index
  print("[State/revert] revert to " .. index)
  revert_state(index)
  local state_text = load_state()
  if state_text then
    state = json.import(state_text)
  else
    state = json.object()
  end
  state_index = index
  return state
end

function state_module.init()
  print("[State/init] loading state")
  local data = load_state()
  if data then
    print("[State/init] state loaded")
    state = json.import(data)
  else
    print("[State/init] no saved state, initializing empty state")
    state = json.object()
  end
  local data = load_store()
  if data then
    print("[State/init] store loaded")
    store = json.import(data)
  else
    print("[State/init] no saved store, initializing empty store")
    store = json.object()
  end
end

function state_module.on_exit()
  print("[State/on_exit] saving state")
  local data = json.export(state)
  save_state(state_index, data)
  print("[State/on_exit] state saved")
end

return state_module
