local user_module = {
  ident = "user",
  need_single = {
    "state",
    "pki",
    "person",
    "node_config"
  },
  provide = {
    api = {},
    user = {}
  }
}

local state

local init_users = {
  admin = json.object{ password = "admin", user_id = "admin", admin = true, home_container = "demo01" },
}

-- ------------------------------------------------------------------------
-- FRONTEND
-- ------------------------------------------------------------------------

local files = {
  [""] = {
    content = [[
      <h2>Module: user</h2>
      <div>Admins:</div>
      <div id="admins"></div>
      <br />
      <div>Users:</div>
      <div id="users"></div>
      <br />
      <button onclick="document.getElementById('add_user_form').classList.toggle('hidden'); return false;">Add user</button>
      <br /><br />
      <form id="add_user_form" class="hidden">
        <div>
          <label>User ID</label>
          <input type="text" name="user_id">
        </div>
        <div>
          <label>Password</label>
          <input type="text" name="password">
        </div>
        <div>
          <label>Home container</label>
          <input type="text" name="home_container">
        </div>
        <div>
          <input type="checkbox" name="admin" value="admin"> <label>Admin</label><br />
          <br />
        </div>
        <div>
          <input type="submit" value="add now" onclick="addUser(this.form); return false;">
        </div>
      </form>
      
      <script>
        function updateUsers() {
          api("POST", "user", { command: "get_users" }, function(result) {
            document.getElementById("admins").innerHTML = "";
            document.getElementById("users").innerHTML = "";
            for (user_id in result) {
              var user = result[user_id];
              var li_el = document.createElement("li");
              console.log(user);
              if (user.identification) {
                if (user.accredited) {
                  li_el.innerHTML = user_id + " (accredited: " + user.identification + ")";
                } else {
                  li_el.innerHTML = user_id + " (not (yet) accredited: " + user.identification + ")";
                }
              } else {
                li_el.innerHTML = user_id + ' (<a href="../person/accredit?person_id=' + user.person_id + '">accredit</a>)';
              }
              if (user.admin) {
                document.getElementById("admins").appendChild(li_el);
              } else {
                document.getElementById("users").appendChild(li_el);
              }
            }
          })
        }
        function addUser(form) {
          api("POST", "user", { command: "add", data: {
            user_id: form.user_id.value,
            password: form.password.value,
            home_container: form.home_container.value,
            admin: form.admin.checked
          } }, function(result) {
            document.getElementById('add_user_form').reset();
            document.getElementById('add_user_form').classList.add('hidden');
            updateUsers()
          });
        }
        updateUsers()
      </script>
    ]]
  },
  ["admin_required"] = {
    content = [[
      <h2>Module: user</h2>
      <div>Admin required</div>
    ]]
  }

}

function user_module.provide.api.on_api_get_file(args)
  print("[user] on api get file '" .. args.file .. "'")
  local user = state.users[args.user_id]
  if not user or not user.admin then
    return files["admin_required"]
  end
  return files[args.file]
end

function user_module.provide.api.on_api_call(args)
  local method = args.method
  local request = args.data
  local command = request.command
  if command == "get_users" then
    local users = json.object()
    for user_id, user in pairs(state.users) do
      users[user_id] = json.object{
        user_id = user_id,
        person_id = user.person_id,
        public_key = user.public_key,
        admin = user.admin
      }
      local person = module.person.get_identification{ person_id = user.person_id }
      if person then
        users[user_id].accredited = person.accredited
        users[user_id].identification = person.identification
      end
    end
    return users
  elseif command == "add" then
    local data = request.data
    
    local key_pair = module.pki.create_key_pair()
    
    state.users[data.user_id] = json.object{
      user_id = data.user_id,
      password = data.password,
      home_container = data.home_container,
      admin = data.admin,
      person_id = key_pair.fingerprint,
      private_key = key_pair.private_key,
      public_key = key_pair.public_key
    }
    
    return json.object{ status = "ok" }
  end
end

-- ------------------------------------------------------------------------
-- AUTHENTICATE WITH PASSWORD
-- ------------------------------------------------------------------------

function user_module.provide.user.authenticate_with_password(args)
  if state.users[args.login] and args.password == state.users[args.login].password then
    return json.object{ 
      user_id = state.users[args.login].user_id,
      person_id = state.users[args.login].person_id,
      admin = state.users[args.login].admin,
      home_container = state.users[args.login].home_container
    }
  end
  return json.object{ message = "Invalid login or password" }
end

-- ------------------------------------------------------------------------
-- SIGN TRANSACTION
-- ------------------------------------------------------------------------

function user_module.provide.user.sign_transaction(args)
  -- TODO don't pass userid, module should get it from environment
  local user = state.users[args.user_id]
  local transaction = args.transaction
  local signature = module.pki.sign{ data = json.export(transaction), private_key = user.private_key }
  return signature.signature
end

function user_module.provide.user.get_public_key(args)
  local user = state.users[args.user_id]
  return user.public_key
end

-- ------------------------------------------------------------------------
-- INIT
-- ------------------------------------------------------------------------

function user_module.init()
  spawn(function()
    print("[User/init] started")
    state = module.state.get_state()
    if not state.users then
      state.users = json.object()
      for user_id, user in pairs(init_users) do
        state.users[user_id] = user
        local key_pair = module.pki.create_key_pair()
        user.person_id = key_pair.fingerprint
        user.private_key = key_pair.private_key
        user.public_key = key_pair.public_key
      end
    end
    print("[User/init] finished")
  end)
end
  
return user_module

